#define INCL_DOSFILEMGR          /* File Manager values */
#define INCL_DOSERRORS           /* DOS Error values    */
#include <os2.h>
#include <builtin.h>
#include <gl.h>
#include <glu.h>
#include <aux.h>
#include <ithread.hpp>

#include "glwindow.hpp"

#define FOVY            20.0
#define MINVIEWDISTANCE 100.0
#define MAXVIEWDISTANCE 200.0

#define ACCUM_NUMBER    9

#ifndef PI
   #define PI 3.14159268
#endif

// GLpaint destructor handler
Boolean GLPaintDestroy::dispatchHandlerEvent( IEvent &event )
{
   Boolean stopDispatching = false;
   switch( event.eventId() )
   {
      case WM_DESTROY:
         stopDispatching = destroy( event );
         break;
   }
   return( stopDispatching );
}

Boolean GLPaintDestroy::destroy( IEvent &event )
{
   // stop the timer
   paintHandler->timer.stop();
   return false;
}

// constructors and destructors
GLPaintHandler::GLPaintHandler( IWindow &_window )
: window( _window ),  paintDestroy( this )
{
   // activate the destructor handler
   paintDestroy.handleEventsFor( &_window );
}

GLPaintHandler::~GLPaintHandler()
{
   HAB hab = IThread::current().anchorBlock();

   // unbind the current context
   pglMakeCurrent( hab, NULL, NULL );

   // destroy the context we created
   pglDestroyContext( hab, hgc );
}

void GLPaintHandler::initOpenGL()
{
   int attribList[] = { PGL_RGBA, PGL_RED_SIZE, 2, PGL_GREEN_SIZE, 2,
      PGL_BLUE_SIZE, 2, PGL_ALPHA_SIZE, 1, PGL_DOUBLEBUFFER, 0 };

   // get a handle to anchor block
   HAB hab =  IThread::current().anchorBlock();

   // create a visual config
   PVISUALCONFIG pVisualConfig = pglChooseConfig( hab, attribList );

   // create a direct OpenGL context
   hgc = pglCreateContext( hab, pVisualConfig, NULL, true );

   // and make the context we got the current one
   Boolean val = pglMakeCurrent( hab, hgc, window.handle() );

   // enable depth buffer
   glEnable( GL_DEPTH_TEST );

   glClearColor( 0.0, 0.0, 0.0, 0.0 );
   glDrawBuffer( GL_FRONT_AND_BACK );
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT  );
   glDrawBuffer( GL_BACK );

   glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
   glShadeModel( GL_FLAT );

   glDepthFunc( GL_LEQUAL );
   glClearAccum( 0.0, 0.0, 0.0, 0.0 );

   // free the visual config created above
   delete [] pVisualConfig;
}

void GLPaintHandler::initLighting()
{
   glEnable( GL_LIGHTING );
   // setup light 0
   glEnable( GL_LIGHT0 );
   glEnable( GL_COLOR_MATERIAL );

   // modify ambient light
   GLfloat amb_light[] = { 0.1, 0.1, 0.1, 1.0 };
   glLightModelfv( GL_LIGHT_MODEL_AMBIENT, amb_light );
}

// openGL settings
void GLPaintHandler::setWindowSize( ISize size )
{
   // make a viewport the acual size of the window
   glViewport( 0, 0, size.width(), size.height() );
}

Boolean GLPaintHandler::paintWindow( IPaintEvent &event )
{
//   GLfloat j8[5][2] = { 0.5, 0.5, 0.3, 0.1, 0.7, 0.9, 0.9, 0.3, 0.1, 0.7 };
   GLfloat j8[9][2] = { 0.5, 0.5, 0.166666, 0.944444, 0.5, 0.166666, 0.5, 0.833333, 0.166666,
               0.277777, 0.833333, 0.388888, 0.166666, 0.611111, 0.833333, 0.722222, 0.833333, 0.055555 };

   // get the size of the window
   GLint viewport[4];
   glGetIntegerv( GL_VIEWPORT, viewport );

   // clear the buffer
   glClear( GL_ACCUM_BUFFER_BIT );

   // no rotations
   glLoadIdentity();

   // place a headlight
   GLfloat light_pos[] = { 0, 0, 0, 1 };
   glLightfv( GL_LIGHT0, GL_POSITION, light_pos );

   for( int jitter=0; jitter<ACCUM_NUMBER; jitter++ )
   {
      // clear the buffer
      glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
      accPerspective( FOVY, (GLdouble)viewport[2] / (GLdouble)viewport[3], MINVIEWDISTANCE, MAXVIEWDISTANCE,
//            j8[jitter][0], j8[jitter][1], 0.0, 0.0, 1.0 );
            0.0, 0.0, 0.33*j8[jitter][0], 0.33*j8[jitter][1], 130.0 );

      // move the image back from the camera
      glTranslatef( 0.0, 0.0, -150.0 );
      glTranslatef( 0.0, -5.0, 0.0 );

      // draw the objects
      glPushMatrix();
      glColor3f( 0.0, 1.0, 0.0 );
      glRotatef( 145, 0.0, 1.0, 1.0 );
      auxSolidCone( 11.0, 30.0 );
      glPopMatrix();

      glPushMatrix();
      glRotatef( 45, 1.0, 1.0, 1.0 );
      glColor3f( 1.0, 0.0, 0.0 );
      auxSolidCube( 20.0 );
      glColor3f( 0.0, 0.0, 1.0 );
      glTranslatef( 0.0, 10.0, 0.0 );
      auxSolidTorus( 5.0, 9.0 );
      glPopMatrix();

      glAccum( GL_ACCUM, 1.0 / ACCUM_NUMBER );
   }
   glAccum( GL_RETURN, 1.0 );

   // swap the back draw buffer to the front
   pglSwapBuffers( IThread::current().anchorBlock(), window.handle() );

   // return true to indicate no additional processing needed
   return( true );
}

// private paint functions
void GLPaintHandler::accFrustum( GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble near,
                        GLdouble far, GLdouble pixdx, GLdouble pixdy, GLdouble eyedx, GLdouble eyedy,
                         GLdouble focus )
{
   // make a frustum for use with the accumulation buffer
   GLdouble xwsize, ywsize;
   GLdouble dx, dy;
   GLint viewport[4];

   glGetIntegerv( GL_VIEWPORT, viewport );
   xwsize = right - left;
   ywsize = top - bottom;
   dx = -(pixdx*xwsize/(GLdouble)viewport[2] + eyedx*near/focus );
   dy = -(pixdx*ywsize/(GLdouble)viewport[3] + eyedy*near/focus );

   glMatrixMode( GL_PROJECTION );
   glLoadIdentity();

   glFrustum( left+dx, right+dx, bottom+dy, top+dy, near, far );
   glMatrixMode ( GL_MODELVIEW );
   glLoadIdentity();
   glTranslatef( -eyedx, -eyedy, 0.0 );
}

void GLPaintHandler::accPerspective( GLdouble fovy, GLdouble aspect, GLdouble near, GLdouble far, GLdouble pixdx,
                        GLdouble pixdy, GLdouble eyedx, GLdouble eyedy, GLdouble focus )
{
   GLdouble fov2, left, right, bottom, top;
   fov2 = ((fovy*PI) / 180.0) / 2.0;

   top = near / (_fcos( fov2 ) / _fsin( fov2 ));
   bottom = -top;
   right = top * aspect;
   left = -right;

   accFrustum( left, right, bottom, top, near, far, pixdx, pixdy, eyedx, eyedy, focus );
}

// window resize handler
Boolean GLResizeHandler::windowResize(IResizeEvent &event )
{
   // tell the paint handler the new size
   window->glPaintHandler().setWindowSize( event.newSize() );
   return false;
}

// constructor
GLWindow::GLWindow()
: IFrameWindow( "OpenGL Demonstration 9", 0x1000, IFrameWindow::defaultStyle()
         | IFrameWindow::noMoveWithOwner | IWindow::synchPaint ),
   canvas( 0x8008, this, this ),
   paintHandler( canvas ),
   resizeHandler( this )
{
   // setup
   setClient( &canvas );
   setDestroyOnClose( true );

   // draw on screen
   sizeTo( ISize( 400, 400 ));
   setFocus();
   show();

   // initialize openGL
   paintHandler.initOpenGL();

   // initialize lighting
   paintHandler.initLighting();

   // set the size of the 3D window
   paintHandler.setWindowSize( canvas.size() );

   // attach the handlers
   paintHandler.handleEventsFor( &canvas );
   resizeHandler.handleEventsFor( &canvas );
}

// destructor
GLWindow::~GLWindow()
{
   // detach the handlers
   paintHandler.stopHandlingEventsFor( &canvas );
   resizeHandler.stopHandlingEventsFor( &canvas );
}
