#define INCL_DOSFILEMGR          /* File Manager values */
#define INCL_DOSERRORS           /* DOS Error values    */
#include <os2.h>
#include <gl.h>
#include <glu.h>
#include <aux.h>
#include <ithread.hpp>

#include "glwindow.hpp"

#define FOVY            20.0
#define MINVIEWDISTANCE 100.0
#define MAXVIEWDISTANCE 200.0

// GLpaint destructor handler
Boolean GLPaintDestroy::dispatchHandlerEvent( IEvent &event )
{
   Boolean stopDispatching = false;
   switch( event.eventId() )
   {
      case WM_DESTROY:
         stopDispatching = destroy( event );
         break;
   }
   return( stopDispatching );
}

Boolean GLPaintDestroy::destroy( IEvent &event )
{
   // stop the timer
   paintHandler->timer.stop();
   return false;
}

// constructors and destructors
GLPaintHandler::GLPaintHandler( IWindow &_window )
: window( _window ),  paintDestroy( this )
{
   // activate the destructor handler
   paintDestroy.handleEventsFor( &_window );
}

GLPaintHandler::~GLPaintHandler()
{
   HAB hab = IThread::current().anchorBlock();

   // unbind the current context
   pglMakeCurrent( hab, NULL, NULL );

   // destroy the context we created
   pglDestroyContext( hab, hgc );
}

void GLPaintHandler::initOpenGL()
{
   int attribList[] = { PGL_RGBA, PGL_RED_SIZE, 2, PGL_GREEN_SIZE, 2,
      PGL_BLUE_SIZE, 2, PGL_ALPHA_SIZE, 1, PGL_DOUBLEBUFFER, 0 };

   // get a handle to anchor block
   HAB hab =  IThread::current().anchorBlock();

   // create a visual config
   PVISUALCONFIG pVisualConfig = pglChooseConfig( hab, attribList );

   // create a direct OpenGL context
   hgc = pglCreateContext( hab, pVisualConfig, NULL, true );

   // and make the context we got the current one
   Boolean val = pglMakeCurrent( hab, hgc, window.handle() );

   glClearColor( 0.0, 0.0, 0.0, 0.0 );
   glDrawBuffer( GL_FRONT_AND_BACK );
   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT  );
   glDrawBuffer( GL_BACK );

   glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
   glShadeModel( GL_FLAT );

   // specify that back facing polygons are to be culled
   glCullFace( GL_BACK );
   glEnable( GL_CULL_FACE );

   // enable polygon antialiasing
   glBlendFunc( GL_SRC_ALPHA_SATURATE, GL_ONE );
   glEnable( GL_BLEND );
   glEnable( GL_POLYGON_SMOOTH );
   glDisable( GL_DEPTH_TEST );

   // free the visual config created above
   delete [] pVisualConfig;
}

void GLPaintHandler::initLighting()
{
   glEnable( GL_LIGHTING );
   // setup light 0
   glEnable( GL_LIGHT0 );
   glEnable( GL_COLOR_MATERIAL );

   // modify ambient light
   GLfloat amb_light[] = { 0.1, 0.1, 0.1, 1.0 };
   glLightModelfv( GL_LIGHT_MODEL_AMBIENT, amb_light );
}

// openGL settings
void GLPaintHandler::setWindowSize( ISize size )
{
   // setup the 3D perspective window
   glMatrixMode( GL_PROJECTION );
   glLoadIdentity();
   gluPerspective( FOVY, (float)window.size().width() / (float)window.size().height(),
                  MINVIEWDISTANCE, MAXVIEWDISTANCE );
   glMatrixMode( GL_MODELVIEW );
   // make a viewport the acual size of the window
   glViewport( 0, 0, size.width(), size.height() );
}

Boolean GLPaintHandler::paintWindow( IPaintEvent &event )
{
   // clear the buffer
   glClear( GL_COLOR_BUFFER_BIT );

   // no rotations
   glLoadIdentity();

   // place a headlight
   GLfloat light_pos[] = { 0, 0, 0, 1 };
   glLightfv( GL_LIGHT0, GL_POSITION, light_pos );

   // move the image back from the camera
   glTranslatef( 0.0, 0.0, -150.0 );
   glTranslatef( 0.0, -5.0, 0.0 );

   // draw the objects
   glPushMatrix();
   glColor3f( 0.0, 1.0, 0.0 );
   glRotatef( 145, 0.0, 1.0, 1.0 );
   auxSolidCone( 11.0, 30.0 );
   glPopMatrix();

   glPushMatrix();
   glRotatef( 45, 1.0, 1.0, 1.0 );
   glColor3f( 1.0, 0.0, 0.0 );
   auxSolidCube( 20.0 );
   glColor3f( 0.0, 0.0, 1.0 );
   glTranslatef( 0.0, 10.0, 0.0 );
   auxSolidTorus( 5.0, 9.0 );
   glPopMatrix();

   // swap the back draw buffer to the front
   pglSwapBuffers( IThread::current().anchorBlock(), window.handle() );

   // return true to indicate no additional processing needed
   return( true );
}

// window resize handler
Boolean GLResizeHandler::windowResize(IResizeEvent &event )
{
   // tell the paint handler the new size
   window->glPaintHandler().setWindowSize( event.newSize() );
   return false;
}

// constructor
GLWindow::GLWindow()
: IFrameWindow( "OpenGL Demonstration 9", 0x1000, IFrameWindow::defaultStyle()
         | IFrameWindow::noMoveWithOwner | IWindow::synchPaint ),
   canvas( 0x8008, this, this ),
   paintHandler( canvas ),
   resizeHandler( this )
{
   // setup
   setClient( &canvas );
   setDestroyOnClose( true );

   // draw on screen
   sizeTo( ISize( 400, 400 ));
   setFocus();
   show();

   // initialize openGL
   paintHandler.initOpenGL();

   // initialize lighting
   paintHandler.initLighting();

   // set the size of the 3D window
   paintHandler.setWindowSize( canvas.size() );

   // attach the handlers
   paintHandler.handleEventsFor( &canvas );
   resizeHandler.handleEventsFor( &canvas );
}

// destructor
GLWindow::~GLWindow()
{
   // detach the handlers
   paintHandler.stopHandlingEventsFor( &canvas );
   resizeHandler.stopHandlingEventsFor( &canvas );
}
